# This file is part of MyPaint.
# Copyright (C) 2009-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2011-2015 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

import logging

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib


logger = logging.getLogger(__name__)


class StateGroup (object):
    """Supervisor instance for GUI states.

    This class mainly deals with the various ways how the user can
    leave such a mode, eg. if the mode is entered by holding down a
    key long enough, it will be left when the key is released.
    """

    def __init__(self):
        super(StateGroup, self).__init__()
        self.states = []
        self.keys_pressed = {}

    @property
    def active_states(self):
        return [s for s in self.states if s.active]

    def create_state(self, enter, leave, popup=None):
        s = State(self, popup)
        s.popup = None  # FIXME: who uses this? hack?
        s.on_enter = enter
        s.on_leave = leave
        self.states.append(s)
        return s

    def create_popup_state(self, popup):
        return self.create_state(popup.enter, popup.leave, popup)


class State (object):
    """A GUI state.

    A GUI state is a mode which the GUI is in, for example an active
    popup window or a special (usually short-lived) view on the
    document. The application defines functions to be called when the
    state is entered or left.
    """

    ## Class consts and instance defaults

    #: How long a key can be held down to go through as single hit (and not
    #: press-and-hold)
    max_key_hit_duration = 0.250

    #: The state is automatically left after this time (ignored during
    #: press-and-hold)
    autoleave_timeout = 0.800

    # : popups only: how long the cursor is allowed outside before closing
    # : (ignored during press-and-hold)"
    #outside_popup_timeout = 0.050

    #: state to activate when this state is activated while already active
    #: (None = just leave this state)
    next_state = None

    #: Allowed buttons and their masks for starting and continuing states
    #: triggered by gdk button press events.
    allowed_buttons_masks = {
        1: Gdk.ModifierType.BUTTON1_MASK,
        2: Gdk.ModifierType.BUTTON2_MASK,
        3: Gdk.ModifierType.BUTTON3_MASK, }

    #: Human-readable display string for the state.
    label = None


    ## Methods

    def __init__(self, stategroup, popup):
        super(State, self).__init__()
        self.sg = stategroup
        self.active = False
        self.popup = popup
        self._autoleave_timeout_id = None
        self._outside_popup_timeout_id = None
        if popup:
            popup.connect("enter-notify-event", self._popup_enter_notify_cb)
            popup.connect("leave-notify-event", self._popup_leave_notify_cb)
            popup.popup_state = self  # FIXME: hacky?
            self.outside_popup_timeout = popup.outside_popup_timeout

    def enter(self, **kwargs):
        logger.debug('Entering State, calling %s', self.on_enter.__name__)
        assert not self.active
        self.active = True
        self._enter_time = Gtk.get_current_event_time()/1000.0
        try:
            self.on_enter(**kwargs)
        except:
            logger.exception("State on_enter method failed")
            raise
        self._restart_autoleave_timeout()

    def leave(self, reason=None):
        logger.debug(
            'Leaving State (reason=%r), calling %s',
            reason,
            self.on_leave.__name__,
        )
        assert self.active
        self.active = False
        self._stop_autoleave_timeout()
        self._stop_outside_popup_timeout()
        self._enter_time = None
        try:
            self.on_leave(reason)
        except:
            logger.exception("State on_leave method failed")
            raise

    def activate(self, action_or_event=None, **kwargs):
        """Activate a State from an action or a button press event.

        :param action_or_event: A Gtk.Action, or a Gdk.Event.
        :param \*\*kwargs: passed to enter().

        For events, only button press events are supported by this code.

        When a Gtk.Action is activated, custom attributes are used to
        figure out whether the action was invoked from a menu, or using
        a keypress.  This requires the action to have been registered
        with the app's keyboard manager.

        See also `keyboard.KeyboardManager.takeover_event()`.

        """
        if self.active:
            # pressing the key again
            if self.next_state:
                self.leave()
                self.next_state.activate(action_or_event)
                return

        # first leave other active states from the same stategroup
        for state in self.sg.active_states:
            state.leave()

        self.keydown = False
        self.mouse_button = None

        if action_or_event:
            if not isinstance(action_or_event, Gtk.Action):
                e = action_or_event
                # eat any multiple clicks. TODO should possibly try to e.g.
                # split a triple click into three clicks in the future.
                if (e.type == Gdk.EventType.DOUBLE_BUTTON_PRESS or
                        e.type == Gdk.EventType.TRIPLE_BUTTON_PRESS):
                    e.type = Gdk.EventType.BUTTON_PRESS

                # currently we only support mouse buttons being single-pressed.
                assert e.type == Gdk.EventType.BUTTON_PRESS
                # let's just note down what mouse button that was
                assert e.button
                if e.button in self.allowed_buttons_masks:
                    self.mouse_button = e.button

            else:
                a = action_or_event
                # register for key release events, see keyboard.py
                if a.keydown:
                    a.keyup_callback = self._keyup_cb
                    self.keydown = True
        self.activated_by_keyboard = self.keydown  # FIXME: should probably be renamed (mouse button possible)
        self.enter(**kwargs)

    def toggle(self, action=None):
        if isinstance(action, Gtk.ToggleAction):
            want_active = action.get_active()
        else:
            want_active = not self.active
        if want_active:
            if not self.active:
                self.activate(action)
        else:
            if self.active:
                self.leave()

    def _keyup_cb(self, widget, event):
        if not self.active:
            return
        self.keydown = False
        if event.time/1000.0 - self._enter_time < self.max_key_hit_duration:
            pass  # accept as one-time hit
        else:
            if self._outside_popup_timeout_id:
                self.leave('outside')
            else:
                self.leave('keyup')

    ## Auto-leave timeout

    def _stop_autoleave_timeout(self):
        if not self._autoleave_timeout_id:
            return
        GLib.source_remove(self._autoleave_timeout_id)
        self._autoleave_timeout_id = None

    def _restart_autoleave_timeout(self):
        if not self.autoleave_timeout:
            return
        self._stop_autoleave_timeout()
        self._autoleave_timeout_id = GLib.timeout_add(
            int(1000*self.autoleave_timeout),
            self._autoleave_timeout_cb,
        )

    def _autoleave_timeout_cb(self):
        if not self.keydown:
            self.leave('timeout')
        return False

    ## Outside-popup timer

    def _stop_outside_popup_timeout(self):
        if not self._outside_popup_timeout_id:
            return
        GLib.source_remove(self._outside_popup_timeout_id)
        self._outside_popup_timeout_id = None

    def _restart_outside_popup_timeout(self):
        if not self.outside_popup_timeout:
            return
        self._stop_outside_popup_timeout()
        self._outside_popup_timeout_id = GLib.timeout_add(
            int(1000*self.outside_popup_timeout),
            self._outside_popup_timeout_cb,
        )

    def _outside_popup_timeout_cb(self):
        if not self._outside_popup_timeout_id:
            return False
        self._outside_popup_timeout_id = None
        if not self.keydown:
            self.leave('outside')
        return False

    def _popup_enter_notify_cb(self, widget, event):
        if not self.active:
            return
        if self._outside_popup_timeout_id:
            GLib.source_remove(self._outside_popup_timeout_id)
            self._outside_popup_timeout_id = None

    def _popup_leave_notify_cb(self, widget, event):
        if not self.active:
            return
        # allow to leave the window for a short time
        if self._outside_popup_timeout_id:
            GLib.source_remove(self._outside_popup_timeout_id)
            self._outside_popup_timeout_id = None
        self._outside_popup_timeout_id = GLib.timeout_add(
            int(1000*self.outside_popup_timeout),
            self._outside_popup_timeout_cb,
        )



# This file is part of MyPaint.
# Copyright (C) 2015 by Andrew Chadwick <a.t.chadwick@gmail.com>
# Copyright (C) 2009 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Non-threaded, prioritizable background processing."""

from __future__ import division, print_function

import collections

from lib.gibindings import GLib


class Processor (object):
    """Queue of low priority tasks for background processing

    Queued tasks are automatically processed in the main thread.
    They run when GTK is idle, or on demand.

    The default priority is much lower than gui event processing.

    """

    def __init__(self, priority=GLib.PRIORITY_LOW):
        """Initialize, specifying a priority"""
        object.__init__(self)
        self._queue = collections.deque()
        self._priority = priority
        self._idle_id = None

    def has_work(self):
        return len(self._queue) > 0

    def add_work(self, func, *args, **kwargs):
        """Adds work

        :param func: a task callable.
        :param *args: passed to func
        :param **kwargs: passed to func

        This starts the queue running if it isn't already.
        Each callable will be called with the given parameters
        until it returns false, at which point it's discarded.

        """
        if not self._idle_id:
            self._idle_id = GLib.idle_add(
                self._process,
                priority=self._priority,
            )
        self._queue.append((func, args, kwargs))

    def finish_all(self):
        """Complete processing: finishes all queued tasks."""
        while self._process():
            pass
        assert self._idle_id is None
        assert len(self._queue) == 0

    def iter_work(self):
        """Iterate across the queued tasks."""
        return iter(self._queue)

    def stop(self):
        """Immediately stop processing and clear the queue."""
        if self._idle_id:
            GLib.source_remove(self._idle_id)
            self._idle_id = None
        self._queue.clear()
        assert self._idle_id is None
        assert len(self._queue) == 0

    def _process(self):
        if not self._idle_id:
            return False
        if len(self._queue) > 0:
            func, args, kwargs = self._queue[0]
            func_done = bool(func(*args, **kwargs))
            if not func_done:
                self._queue.popleft()
        if len(self._queue) == 0:
            self._idle_id = None
        return bool(self._queue)



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2017 by the MyPaint Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Bulk management of layer visibility: presentation layer and view."""

# Imports:

from __future__ import print_function, division

import gui.mvp
import logging
import gui.dialogs
import gui.application
import lib.layervis
from lib.gettext import C_
from lib.xml import escape

from lib.gibindings import Gtk


# Module vars:

logger = logging.getLogger(__name__)


# Class definitions:

class LayerViewUI (gui.mvp.BuiltUIPresenter, object):
    """Presents the document's named layer views in a row of controls."""

    def __init__(self, docmodel):
        super(LayerViewUI, self).__init__()
        self._docmodel = docmodel
        self._lvm = docmodel.layer_view_manager
        self._app = gui.application.get_app()

    def _init_model(self):
        lvm = self._lvm
        lvm.current_view_changed += self._current_view_changed_cb
        lvm.view_names_changed += self._view_names_changed_cb

    def init_view(self):
        self._init_model()
        store = Gtk.ListStore(str, str)   # columns: <our_id, display_markup>
        markup = C_(
            "view controls: dropdown: item markup",
            u"<i>{builtin_view_name}</i>",
        ).format(
            builtin_view_name=escape(lib.layervis.UNSAVED_VIEW_DISPLAY_NAME),
        )
        # None has a special meaning for GTK combo ID columns,
        # so we substitute the empty string. Corrolory: you can't name
        # views to the empty string.
        store.append([u"", markup])
        store.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self._store = store

        combo = self.view.current_view_combo
        combo.set_model(store)
        combo.set_id_column(0)

        cell = self.view.layer_text_cell
        combo.add_attribute(cell, "markup", 1)

        self._refresh_ui()

    # MVP implementation:

    @property
    def widget(self):
        return self.view.layout_grid

    # View updater callbacks:

    @gui.mvp.view_updater
    def _current_view_changed_cb(self, lvm):
        self._update_current_view_name()
        self._update_buttons()

    @gui.mvp.view_updater
    def _view_names_changed_cb(self, lvm):
        self._update_view_combo_names()

    @gui.mvp.view_updater
    def _refresh_ui(self):
        self._update_view_combo_names()
        self._update_current_view_name()
        self._update_buttons()

    @gui.mvp.view_updater
    def _update_view_combo_names(self):
        """Update the combo box to reflect a changed list names."""
        model_names = set(self._lvm.view_names)
        store = self._store

        kept_names = set()
        it = store.get_iter_first()
        while it is not None:
            name, = store.get(it, 0)
            if isinstance(name, bytes):
                name = name.decode("utf-8")
            # Skip over our representation of the builtin
            if name == u"":
                it = store.iter_next(it)
                continue
            # Drop or keep
            if name not in model_names:
                it_still_valid = store.remove(it)  # advances it
                if not it_still_valid:
                    it = None
            else:
                it = store.iter_next(it)
                kept_names.add(name)

        for name in model_names:
            if name not in kept_names:
                markup = C_(
                    "view controls: dropdown: item markup",
                    u"{user_chosen_view_name}",
                ).format(
                    user_chosen_view_name=escape(name),
                )
                store.append([name, markup])

    @gui.mvp.view_updater
    def _update_buttons(self):
        current = self._lvm.current_view_name
        op_buttons = [
            self.view.rename_button,
            self.view.remove_button,
            self.view.locked_button,
        ]
        for b in op_buttons:
            b.set_sensitive(current is not None)

        icon_name = "mypaint-object-unlocked-symbolic"
        active = self._lvm.current_view_locked
        if self._lvm.current_view_locked:
            icon_name = "mypaint-object-locked-symbolic"
        self.view.locked_image.set_from_icon_name(
            icon_name,
            Gtk.IconSize.BUTTON,
        )
        btn = self.view.locked_button
        if bool(active) != bool(btn.get_active()):
            btn.set_active(active)

    @gui.mvp.view_updater
    def _update_current_view_name(self):
        combo = self.view.current_view_combo
        id_ = self._lvm.current_view_name
        if id_ is None:
            id_ = ""
        combo.set_active_id(id_)

    # Model updaters:

    @gui.mvp.model_updater
    def _add_button_clicked_cb(self, button):
        doc = self._docmodel
        name = self._lvm.current_view_name
        if name is None:
            name = lib.layervis.NEW_VIEW_IDENT
        cmd = lib.layervis.AddLayerView(doc, name=name)
        doc.do(cmd)

    @gui.mvp.model_updater
    def _remove_button_clicked_cb(self, button):
        doc = self._docmodel
        cmd = lib.layervis.RemoveActiveLayerView(doc)
        doc.do(cmd)

    @gui.mvp.model_updater
    def _rename_button_clicked_cb(self, button):
        old_name = self._lvm.current_view_name
        if old_name is None:
            return
        new_name = gui.dialogs.ask_for_name(
            title=C_(
                "view controls: rename view dialog",
                u"Rename View",
            ),
            widget=self._app.drawWindow,
            default=old_name,
        )
        if new_name is None:
            return
        new_name = new_name.strip()
        if new_name == old_name:
            return
        doc = self._docmodel
        cmd = lib.layervis.RenameActiveLayerView(doc, new_name)
        doc.do(cmd)

    @gui.mvp.model_updater
    def _locked_button_toggled_cb(self, button):
        if self._lvm.current_view_name is None:
            return
        new_locked = button.get_active()
        doc = self._docmodel
        cmd = lib.layervis.SetActiveLayerViewLocked(doc, locked=new_locked)
        doc.do(cmd)

    @gui.mvp.model_updater
    def _current_view_combo_changed_cb(self, combo):
        name = combo.get_active_id()
        if name is None:
            return  # specifically means "nothing active" in GTK
        if name == "":
            name = None
        if name == self._lvm.current_view_name:
            return  # nothing to do
        doc = self._docmodel
        cmd = lib.layervis.ActivateLayerView(doc, name)
        doc.do(cmd)



# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Deprecated interface for tabbed color selectors.

Newly written color adjusters should present the attributes described in
gui.workspace instead of implementing this interface.

"""

from __future__ import division, print_function

from .adjbases import ColorAdjuster


class CombinedAdjusterPage (ColorAdjuster):
    """Interface for multitab page content.

    Page instances are expected to distribute `set_color_manager()` to each of
    their component controls, and also to implement the methods defined in this
    interface.

    """

    @classmethod
    def get_page_icon_name(class_):
        """Returns the page's icon name.
        """
        raise NotImplementedError

    @classmethod
    def get_page_title(class_):
        """Returns the title for the page.

        Word as "this page/tab contains a [...]", in titlecase.
        """
        raise NotImplementedError

    @classmethod
    def get_page_description(class_):
        """Returns the descriptive text for the page.

        Word as "this page/tab lets you [...]", in titlecase.
        """
        raise NotImplementedError

    @classmethod
    def get_properties_description(class_):
        """Override & return a string if `show_properties()` is implemented.

        The returned string should explain what the properties button does. The
        default implemented here returns None, which also indicates that no
        properties dialog is implemented.
        """
        return None

    def show_properties(self):
        """Override to show the page's properties dialog.
        """
        pass

    def get_page_widget(self):
        """Returns the `Gtk.Table` instance for the page body.
        """
        raise NotImplementedError
